/*
 * vfat.c -- probe for label and serial number of fat/vfat filesystems
 *           (this is largely a chop-job from libblkid in e2fsprogs)
 *
 * Copyright (C) 1999 by Andries Brouwer
 * Copyright (C) 1999, 2000, 2003 by Theodore Ts'o
 * Copyright (C) 2001 by Andreas Dilger
 * Copyright 2007 Red Hat, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 */

#define _GNU_SOURCE 1
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <endian.h>

#include "vfat.h"

static inline  uint16_t vfat_swab16(uint16_t val)
{
	return (val >> 8) | (val << 8);
}

static inline uint32_t vfat_swab32(uint32_t val)
{
	return ((val>>24) | ((val>>8)&0xFF00) |
		((val<<8)&0xFF0000) | (val<<24));
}

static inline uint64_t vfat_swab64(uint64_t val)
{
	return (vfat_swab32(val >> 32) |
		(((uint64_t) vfat_swab32(val & 0xFFFFFFFFUL)) << 32));
}

#if  __BYTE_ORDER == __BIG_ENDIAN
#define vfat_le16(x) vfat_swab16(x)
#define vfat_le32(x) vfat_swab32(x)
#define vfat_le64(x) vfat_swab64(x)
#define vfat_be16(x) (x)
#define vfat_be32(x) (x)
#define vfat_be64(x) (x)
#else
#define vfat_le16(x) (x)
#define vfat_le32(x) (x)
#define vfat_le64(x) (x)
#define vfat_be16(x) vfat_swab16(x)
#define vfat_be32(x) vfat_swab32(x)
#define vfat_be64(x) vfat_swab64(x)
#endif

static unsigned char *get_buffer(int fd,
				 off_t pos,
				 size_t len)
{
	unsigned char *newbuf = NULL;
	off_t old_pos;

	old_pos = lseek(fd, 0, SEEK_CUR);

	if (lseek(fd, pos, SEEK_SET) < 0)
		return NULL;

	newbuf = calloc(1, len);
	if (!newbuf) {
		lseek(fd, old_pos, SEEK_CUR);
		return NULL;
	}
	if (read(fd, newbuf, len) < 0) {
		free(newbuf);
		newbuf = NULL;
	}

	lseek(fd, old_pos, SEEK_CUR);
	return newbuf;
}

#define FAT_ATTR_VOLUME_ID		0x08
#define FAT_ATTR_DIR			0x10
#define FAT_ATTR_LONG_NAME		0x0f
#define FAT_ATTR_MASK			0x3f
#define FAT_ENTRY_FREE			0xe5

static char *no_name = "NO NAME    ";

static unsigned char *search_fat_label(struct vfat_dir_entry *dir, int count)
{
	unsigned int i;

	for (i = 0; i < count; i++) {
		if (dir[i].name[0] == 0x00)
			break;
		
		if ((dir[i].name[0] == FAT_ENTRY_FREE) ||
		    (dir[i].cluster_high != 0 || dir[i].cluster_low != 0) ||
		    ((dir[i].attr & FAT_ATTR_MASK) == FAT_ATTR_LONG_NAME))
			continue;

		if ((dir[i].attr & (FAT_ATTR_VOLUME_ID | FAT_ATTR_DIR)) == 
		    FAT_ATTR_VOLUME_ID) {
			return dir[i].name;
		}
	}
	return 0;
}

static int figure_label_len(const unsigned char *label, int len)
{
	const unsigned char *end = label + len - 1;

	while ((*end == ' ' || *end == 0) && end >= label)
		--end;
	if (end >= label) {
		label = label;
		return end - label + 1;
	}
	return 0;
}

static inline uint16_t copy_u8_to_u16(uint8_t *in)
{
	uint16_t ret;
	uint8_t *buf = (void *)&ret;

	buf[0] = in[0];
	buf[1] = in[1];
	return ret;
}

/* FAT label extraction from the root directory taken from Kay
 * Sievers's volume_id library */
static int probe_fat(
        int fd,
        char *buf,
        char **label,
        char *uuid)
{
	struct vfat_super_block *vs = (struct vfat_super_block *) buf;
	struct msdos_super_block *ms = (struct msdos_super_block *) buf;
	struct vfat_dir_entry *dir = NULL;
	unsigned char *vol_label = 0;
	unsigned char *vol_serno;
	int label_len = 0, maxloop = 100;
	uint16_t sector_size, dir_entries, reserved;
	uint32_t sect_count, fat_size, dir_size, cluster_count, fat_length;
	uint32_t buf_size, start_data_sect, next, root_start, root_dir_entries;

	/* sector size check */
	sector_size = vfat_le16(copy_u8_to_u16(ms->ms_sector_size));
	if (sector_size != 0x200 && sector_size != 0x400 &&
	    sector_size != 0x800 && sector_size != 0x1000)
		return 1;

	dir_entries = vfat_le16(copy_u8_to_u16(ms->ms_dir_entries));
	reserved =  vfat_le16(ms->ms_reserved);
	sect_count = vfat_le16(copy_u8_to_u16(ms->ms_sectors));
	if (sect_count == 0)
		sect_count = vfat_le32(ms->ms_total_sect);

	fat_length = vfat_le16(ms->ms_fat_length);
	if (fat_length == 0)
		fat_length = vfat_le32(vs->vs_fat32_length);

	fat_size = fat_length * ms->ms_fats;
	dir_size = ((dir_entries * sizeof(struct vfat_dir_entry)) +
			(sector_size-1)) / sector_size;

	cluster_count = sect_count - (reserved + fat_size + dir_size);
	cluster_count /= ms->ms_cluster_size;

	if (cluster_count > FAT32_MAX)
		return 1;

	if (ms->ms_fat_length) {
		/* the label may be an attribute in the root directory */
		root_start = (reserved + fat_size) * sector_size;
		root_dir_entries = vs->vs_dir_entries[0] + 
			(vs->vs_dir_entries[1] << 8);

		buf_size = root_dir_entries * sizeof(struct vfat_dir_entry);
		dir = (struct vfat_dir_entry *) get_buffer(fd, root_start,
							  buf_size);
		if (dir)
			vol_label = search_fat_label(dir, root_dir_entries);

		if (!vol_label || !memcmp(vol_label, no_name, 11))
			vol_label = ms->ms_label;

		if (vol_label) {
			label_len = figure_label_len(ms->ms_label, 11);
			vol_label = strndup(vol_label, label_len);
		}
		
		if (dir) {
			free(dir);
                        dir = NULL;
                }

		vol_serno = ms->ms_serno;
	} else {
		/* Search the FAT32 root dir for the label attribute */
		buf_size = vs->vs_cluster_size * sector_size;
		start_data_sect = reserved + fat_size;

		next = vfat_le32(vs->vs_root_cluster);
		while (next && --maxloop) {
			uint32_t next_sect_off;
			uint64_t next_off, fat_entry_off;
			int count;

			next_sect_off = (next - 2) * vs->vs_cluster_size;
			next_off = (start_data_sect + next_sect_off) * 
				sector_size;

			dir = (struct vfat_dir_entry *) 
				get_buffer(fd, next_off, buf_size);
			if (dir == NULL)
				break;

			count = buf_size / sizeof(struct vfat_dir_entry);

			vol_label = search_fat_label(dir, count);
			if (vol_label) {
				label_len = figure_label_len(vol_label, 11);
				vol_label = strndup(vol_label, label_len);
				free(dir);
                                dir = NULL;
				break;
			}
			free(dir);
                        dir = NULL;

			/* get FAT entry */
			fat_entry_off = (reserved * sector_size) + 
				(next * sizeof(uint32_t));
			buf = get_buffer(fd, fat_entry_off, buf_size);
			if (buf == NULL)
				break;

			/* set next cluster */
			next = vfat_le32(*((uint32_t *) buf) & 0x0fffffff);
			free(buf);
		}

		if (vol_label && !memcmp(vol_label, no_name, 11)) {
			free(vol_label);
			vol_label = NULL;
		}
			
		if (!vol_label) {
			label_len = figure_label_len(vs->vs_label, 11);
			vol_label = strndup(vs->vs_label, label_len);
		}
		vol_serno = vs->vs_serno;
	}

	if (vol_label && memcmp(vol_label, no_name, 11)) {
		if ((label_len = figure_label_len(vol_label, 11)))
			*label = vol_label;
	}

	/* We can't just print them as %04X, because they are unaligned */
        snprintf(uuid, 16, "%02X%02X-%02X%02X", vol_serno[3], vol_serno[2],
		vol_serno[1], vol_serno[0]);

	if (dir)
		free(dir);

	return 0;
}

/*
 * The FAT filesystem could be without a magic string in superblock
 * (e.g. old floppies).  This heuristic for FAT detection is inspired
 * by http://vrfy.org/projects/volume_id/ and Linux kernel.
 * [7-Jul-2005, Karel Zak <kzak@redhat.com>]
 */
static int probe_fat_nomagic(int fd,
        char *buf,
        char **label,
        char *uuid)
{
	struct vfat_super_block *vs;

	vs = (struct vfat_super_block *)buf;

	/* heads check */
	if (vs->vs_heads == 0)
		return 1;

	/* cluster size check*/
	if (vs->vs_cluster_size == 0 ||
	    (vs->vs_cluster_size & (vs->vs_cluster_size-1)))
		return 1;

	/* fat counts(Linux kernel expects at least 1 FAT table) */
	if (!vs->vs_fats)
		return 1;

	return probe_fat(fd, buf, label, uuid);
}

struct vfat_blkid_magic;

typedef int (*vfat_blkid_probe_t)(int fd, char *buf, char **label, char *uuid);

struct vfat_blkid_magic {
	const char	*bim_type;	/* type name for this magic */
	long		bim_kboff;	/* kilobyte offset of superblock */
	unsigned	bim_sboff;	/* byte offset within superblock */
	unsigned	bim_len;	/* length of magic */
	const char	*bim_magic;	/* magic string */
	vfat_blkid_probe_t	bim_probe;	/* probe function */
};

/*
 * VFAT_BLKID_BLK_OFFS is at least as large as the highest bim_kboff defined
 * in the type_array table below + bim_kbalign.
 *
 * When probing for a lot of magics, we handle everything in 1kB buffers so
 * that we don't have to worry about reading each combination of block sizes.
 */
#define VFAT_BLKID_BLK_OFFS	0	/* currently vfat */

/*
 * Various filesystem magics that we can check for.  Note that kboff and
 * sboff are in kilobytes and bytes respectively.  All magics are in
 * byte strings so we don't worry about endian issues.
 */
static struct vfat_blkid_magic type_array[] = {
  { "vfat",      0,   0x52,  5, "MSWIN",                probe_fat },
  { "vfat",      0,   0x52,  8, "FAT32   ",             probe_fat },
  { "vfat",      0,   0x36,  5, "MSDOS",                probe_fat },
  { "vfat",      0,   0x36,  8, "FAT16   ",             probe_fat },
  { "vfat",      0,   0x36,  8, "FAT12   ",             probe_fat },
  { "vfat",      0,      0,  2, "\353\220",             probe_fat_nomagic },
  { "vfat",      0,      0,  1, "\351",                 probe_fat_nomagic },
  {   NULL,	 0,	 0,  0, NULL,			NULL }
};

int get_vfat_label_uuid(int fd, char **label, char *uuid)
{
        int idx;
        struct vfat_blkid_magic *id;
	unsigned char buf[1024];

	for (id = type_array; id->bim_type; id++) {
                memset(buf, '\0', sizeof (buf));
		idx = id->bim_kboff + (id->bim_sboff >> 10);
		if (idx > VFAT_BLKID_BLK_OFFS || idx < 0)
			continue;

                if (lseek(fd, idx << 10, SEEK_SET) < 0)
                    continue;

                if (read(fd, buf, 1024) != 1024)
                    continue;

		if (memcmp(id->bim_magic, buf + (id->bim_sboff&0x3ff),
			   id->bim_len)) {
			continue;
                }

		if ((id->bim_probe == NULL) ||
		    (id->bim_probe(fd, buf, label, uuid) == 0)) {
                        return 1;
		}
	}
        return 0;
}
